#!/usr/bin/env python
#
# evaluate.py - Pyfeeds test evaluation logic
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains the :func:`evaluateTest` function, and other functions
used for evaluating the results of a pyfeeds test.


All available evaluation routines are defined in this module, and are
identified by their name - any function with a name that begins in ``eval``,
but not ``evaluate``, is considered to be an evaluation routine.


All evaluation routines must accept two parameters, which are the names of two
files to be compared.


Evaluation routines ending with the word ``Group`` are routines which accept
two groups of files, and evaluate the groups as a whole.


All evaluation routines must return a numeric value, where 0 indicates that
the files are identical, and non-0 indicates that the files are different.
"""

from __future__ import print_function
from __future__ import division

import              logging
import              fnmatch
import              glob
import              sys
import              os
import os.path   as op
import functools as ft

import numpy     as np
import nibabel   as nib

from . import common
from . import hashing


log = logging.getLogger(__name__)


def compareDirs(pyf):
    """The ``compare`` command. Compares the ``inputDir`` against the
    ``benchmarkDir`` using the :func:`compareAllFiles` function.
    """
    compareAllFiles(pyf, None, pyf.inputDir, pyf.benchmarkDir, pyf.outputFile)


def listRoutines():
    """Returns a list containing the names of all available evaluation
    routines.
    """

    thismod  = sys.modules[__name__]
    routines = []

    for attr in thismod.__dict__.keys():
        if attr.startswith('eval') and not attr.startswith('evaluate'):
            routines.append(attr)

    return routines


def isGroupRoutine(rname):
    """Returns ``True`` if the specified routine is a file group evaluation
    routine, ``False``otherwise.
    """
    return rname.endswith('Group')


def evaluateTest(pyf, test, retcode, outputDir, benchmarkDir):
    """Evaluate the given :class:`.Test` instance.

    The following rules are used in evaluating a test:

    1. If the test returns non-0, then it fails.

    2. If the test is in the ``selfEval`` list, then it passes or fails based
       on its return code.

    3. If there is no benchmark data for the test, then it passes or fails
       based on its return code.

    4. Otherwise, the test is evaluated by comparing every file in its
       benchmark data with the corresponding file in the test output.
       This is done by the :func:`compareAllFiles` function.

    :arg pyf:          The :class:`.Pyfeeds`
    :arg test:         The :class:`.Test` instance to be evaluated
    :arg retcode:      Return code of the test process
    :arg outputDir:    Directory containing the test output
    :arg benchmarkDir: Directory containing the benchmark data for the test
                       (or ``None`` if this test has no benchmark data).
    """

    # Rule 1: If a test returns non-0, that
    #         means it failed
    if retcode != 0:
        log.debug('Test {} returned non-zero - failing'.format(test.name))
        return False

    # Rule 2: If the test is in the self_eval
    #         list, this means that it evaluated
    #         itself. It eturned 0, so it passes.
    if isInSelfEval(pyf, test):
        log.debug('Test {} is in selfEval - passing'.format(test.name))
        return True

    if benchmarkDir is None:
        log.debug('Test {} has no benchmark data - passing'.format(test.name))
        return True

    # Rule 4. For every file in the benchmark
    #         data, make sure there is a
    #         corresponding file in the test
    #         output, and make sure it matches
    outFile = op.join(outputDir, 'feedsResults.log')
    log.debug('Test {}: evaluating against benchmark data'.format(test.name))
    return compareAllFiles(pyf, test, outputDir, benchmarkDir, outFile)


def getEvaluationRoutines(pyf, filename):
    """Searches for evaluation routines suitable for testing the given
    ``filename``.

    :returns: A list of tuples, each containing:

                - An evaluation routine function
                - The name of the function

              Or ``[]`` if there are no routines to test the file.
    """

    thismod = sys.modules[__name__]
    matches = {}

    for pattern, rnames in pyf.evalRoutines.items():
        if fnmatch.fnmatch(filename, pattern):
            rfuncs           = [getattr(thismod, n) for n in rnames]
            rfuncs           = [ft.update_wrapper(ft.partial(f, pyf), f)
                                for f in rfuncs]
            routines         = list(zip(rnames, rfuncs))
            matches[pattern] = routines

    # if the file matched multiple patterns,
    # use the longest one, as it is probably
    # the most specific
    routines = []
    if len(matches) > 0:
        routines = matches[max(matches, key=len)]

    return routines


def getFileGroup(pyf, filename):
    """Determines whether the specified file is part of a file group.

    :returns: A tuple containing:

                - A list of all files that are in the group

                - The file patterns in the matched group, or ``None`` if the
                  file is not part of a group.
    """

    filename = op.abspath(filename)
    group    = None

    # find a file group which
    # matches the given file
    for g in pyf.fileGroups:
        for pat in g:

            # All files in a group must
            # be in the same directory
            if fnmatch.fnmatch(op.basename(filename), pat):
                group = g
                break

    if group is None:
        return [filename], None

    dirname  = op.dirname(filename)
    allFiles = []

    for pat in group:
        hits = glob.glob(op.join(dirname, pat))

        for hit in hits:
            hit = op.abspath(hit)
            if hit not in allFiles:
                allFiles.append(hit)

    return list(sorted(allFiles)), tuple(group)


def isInSelfEval(pyf, test):
    """Returns ``True`` if the given :class:`.Test` instance is in the
    ``selfEval`` list, indicating that it has evaluated itself.
    """
    testDir = test.testDir
    matches = [testDir.endswith(p) for p in pyf.selfEval]
    return any(matches)


def isInExcludeList(pyf, filename):
    """Returns ``True`` if the given filename matches any patterns
    in the ``exclude`` list, indicating that it should not be evaluated.
    """
    matches = [fnmatch.fnmatch(filename, p) for p in pyf.exclude]
    return any(matches)


def getTolerance(pyf, filename):
    """Returns an error tolerance to use when evaluating the given
    ``filename`` - either a tolerance in the ``tolerances`` list, or
    the ``defaultTolerance``.
    """
    for pattern, tol in pyf.tolerances.items():
        if fnmatch.fnmatch(filename, pattern):
            return tol

    return pyf.defaultTolerance


def compareAllFiles(pyf, test, outputDir, benchmarkDir, logFile=None):
    """Evaluate the files generated by the given :class:`.Test` instance.

    Walks through the ``benchmarkDir`` comparing every file in it with the
    corresponding file in the ``outputDir``.

    If ``logFile`` is not provided, the output is printed to standard out.
    """

    results         = []
    finalResult     = True
    benchmarks      = []
    groupsEvaluated = {}

    for base, dirs, filenames in os.walk(benchmarkDir):
        filenames = [op.join(base, f) for f in filenames]
        benchmarks.extend(filenames)

    # Loop through every file in the benchmark directory
    for benchmark in benchmarks:

        relbenchmark      = op.relpath(benchmark, benchmarkDir)
        testfile          = op.join(outputDir, relbenchmark)
        routines          = getEvaluationRoutines(pyf, benchmark)
        tolerance         = getTolerance(pyf, benchmark)
        groupfiles, group = getFileGroup(pyf, benchmark)

        # Skip files which match a
        # pattern in the exclude list
        if isInExcludeList(pyf, benchmark):
            results.append((relbenchmark, 'EXCLUDE', None, None, None))
            continue

        # Skip files which do not match
        # a pattern in the evalRoutines
        # list
        if len(routines) == 0:
            results.append((relbenchmark, 'SKIP', None, None, None))
            continue

        # Fail files which exist in the
        # benchmark data, but do not exist
        # in the test output
        if not op.exists(testfile):
            results.append((relbenchmark, 'FAIL', 'file_missing', 0, 0))
            continue

        # Otherwise, test the file. If
        # any evaluation routine generates
        # a value above the tolerance,
        # the test fails for this file.
        try:

            for rname, rfunc in routines:

                # if this is a file group evaluation
                # routine, retrieve all other files
                # that are part of the group
                if isGroupRoutine(rname):

                    # this file group has already
                    # been evaluated. Skip the
                    # evaluation routine, and return
                    # the previously calculated
                    # result.
                    if (rname, group) in groupsEvaluated:

                        def rfunc_cached(*a):
                            error = groupsEvaluated[rname, group]
                            log.debug('Using saved group evaluation '
                                      'results (%s) for %s (group: %s): '
                                      '%0.2f', rname, relbenchmark, group,
                                      error)
                            return error
                        rfunc = ft.update_wrapper(rfunc_cached, rfunc)

                    rargs = [getFileGroup(pyf, testfile)[0], groupfiles]

                else:
                    rargs = [testfile, benchmark]

                error       = rfunc(*rargs)
                result      = error <= tolerance
                finalResult = finalResult and result

                log.debug('%s: %s - %s', testfile, rfunc.__name__, error)

                if result: status = 'PASS'
                else:      status = 'FAIL'

                # Cache results for file group evaluations,
                # so we don't unnecessarily re-evaluate them
                if isGroupRoutine(rname) and \
                   (rname, group) not in groupsEvaluated:
                    log.debug('Saving group evaluation results (%s) '
                              'for %s (group: %s): %0.2f',
                              rname, relbenchmark, group, error)
                    groupsEvaluated[rname, group] = error

                results.append((relbenchmark, status, rname, error, tolerance))

        except Exception:
            log.error('Error evaluating {}'.format(relbenchmark),
                      exc_info=True)
            results.append((relbenchmark, 'ERROR', rname, None, tolerance))
            finalResult = False

    # Every file that gets tested is added to
    # a log file with a CSV format:
    #
    # file, status, routineName, error, tolerance
    def fmtlog(fname, status, routine, error, tolerance):
        if routine   is None: routine   = 'none'
        if error     is None: error     = np.nan
        if tolerance is None: tolerance = np.nan
        fields = [fname,
                  status,
                  routine,
                  '{:0.6g}'.format(error),
                  '{:0.6g}'.format(tolerance)]

        return '{}'.format(','.join(fields))

    def sortkey(result):
        status = result[1]
        error  = result[3]

        statuswt = {'ERROR'   : 500,
                    'FAIL'    : 100,
                    'PASS'    : 50,
                    'SKIP'    : 10,
                    'EXCLUDE' : 1}[status]

        if error is None:
            error = 1
        return statuswt + error

    lines = ['FILE,RESULT,ROUTINE,ERROR,TOLERANCE']
    # sort by error value
    for result in reversed(sorted(results, key=sortkey)):
        lines.append(fmtlog(*result))

    if logFile is not None:
        with open(logFile, 'wt') as f:
            for line in lines:
                f.write(line + '\n')
    else:
        for line in lines:
            print(line)

    return finalResult


def evalHeader(pyf, testfile, benchmark, alldims=True):
    """Evaluation routine which compares the header fields of two NIFTI
    images.

    Returns 0 if they all match, 1 otherwise.
    """

    img1 = pyf.imageCache[testfile]
    img2 = pyf.imageCache[benchmark]

    hdr1   = img1.header
    hdr2   = img2.header
    fields = ['dim',       'pixdim',     'intent_code',
              'datatype',  'qform_code', 'sform_code',
              'quatern_b', 'quatern_c',  'quatern_d',
              'qoffset_x', 'qoffset_y',  'qoffset_z',
              'srow_x',    'srow_y',     'srow_z']

    for f in fields:
        f1 = hdr1[f]
        f2 = hdr2[f]

        if (not alldims) and (f in ('dim', 'pixdim')):
            ndim = img1.header['dim'][0]
            f1   = f1[:ndim + 1]
            f2   = f2[:ndim + 1]

        if not np.all(np.isclose(f1, f2)):
            return 1

    return 0


def evalHeaderRestrictDims(pyf, testfile, benchmark):
    """Evaluation routine which compares the header fields of two NIFTI
    images. For the `dim` and `pixdim` fields, only the entries which
    are expected to be valid (e.g. `dim1`, `dim2`, and `dim3` for a 3D image)
    are compared.

    Returns 0 if they all match, 1 otherwise.
    """
    return evalHeader(pyf, testfile, benchmark, alldims=False)


def evalImage(pyf, testfile, benchmark):
    """Evaluation routine which compares the data from two NIFTI images.

    The :func:`cmpArrays` function does the calculation.
    """

    img1  = pyf.imageCache[testfile]
    img2  = pyf.imageCache[benchmark]
    data1 = img1.get_data()
    data2 = img2.get_data()

    return cmpArrays(data1, data2)


def evalNumericalText(pyf, testfile, benchmark):
    """Evaluation routine which compares the numerical data from two text
    files.

    The :func:`cmpArrays` function does the calculation.
    """

    data1 = common.loadNumericText(testfile)
    data2 = common.loadNumericText(benchmark)

    return cmpArrays(data1, data2)


def evalMD5(pyf, testfile, benchmark):
    """Compares the given files by calculating their MD5 hash. Returns
    0 if the hashes match, 1 otherwise.
    """

    hash1 = hashing.hashFile(testfile)
    hash2 = hashing.hashFile(benchmark)

    if hash1 != hash2: return 1
    else:              return 0


def evalVectorImage(pyf, testfile, benchmark):
    """Compares the given NIFTI images. It is assumed that they are
    both vector images, where each voxel contains a 3-dimensional
    vector, undirected, and centered at 0.
    """
    pion2 = np.pi / 2
    img1  = pyf.imageCache[testfile]
    img2  = pyf.imageCache[benchmark]

    data1 = img1.get_data().reshape(-1, 3).T
    data2 = img2.get_data().reshape(-1, 3).T

    # Calculate the length of each vector,
    # discard vectors of length 0, and
    # normalise each vector to unit length
    len1  = np.linalg.norm(data1, axis=0)
    len2  = np.linalg.norm(data2, axis=0)
    nz1   = len1 > 1e-6
    nz2   = len2 > 1e-6
    nz    = nz1 & nz2

    data1 = data1[:, nz] / len1[nz]
    data2 = data2[:, nz] / len2[nz]
    len1  = len1[nz]
    len2  = len2[nz]

    # Calculate the angle between each vector.
    # Vectors are undirected, and centered at
    # (0, 0, 0), so the maximum possible angle
    # we can have is 90 degrees.
    dot          = np.sum(data1 * data2, axis=0)
    dot          = np.clip(dot, -1, 1)
    angle        = np.arccos(dot)
    amask        = angle > pion2
    angle[amask] = np.pi - angle[amask]

    # We also compare the length of each
    # vector, and the pattern of missing
    # voxels (vectors of length 0)
    nzcorr  = np.abs(np.corrcoef(nz1, nz2)[0, 1])
    lendiff = np.abs(len1 - len2) / np.max((len1, len2), axis=0)
    angle   = np.abs(angle) / pion2

    if np.isnan(nzcorr):
        nzcorr = 1

    # All errors are normalised to
    # the range (0, 1). We return
    # the worst error
    nzError    = 1 - nzcorr
    angleError = angle.mean()
    lenError   = lendiff.mean()

    return max((nzError, lenError, angleError))


def evalPolarCoordinateImageGroup(pyf, testfiles, benchmarks):
    """Compares the values in a set of files containing 3D polar coordinates.

    Currently only the ``phi`` and ``theta`` outputs of ``bedpostx`` are
    understood.

    The coordinates are converted into vectors, and then passed to the
    :func:`evalVectorImage` function.
    """

    thetapatterns = [
        'mean_th?samples.*',
        'merged_th?samples.*',
    ]

    phipatterns = [
        'mean_ph?samples.*',
        'merged_ph?samples.*',
    ]

    def matches(fname, patterns):
        return any([fnmatch.fnmatch(op.basename(fname), p) for p in patterns])

    testtheta  = [f for f in testfiles  if matches(f, thetapatterns)]
    testphi    = [f for f in testfiles  if matches(f, phipatterns)]
    benchtheta = [f for f in benchmarks if matches(f, thetapatterns)]
    benchphi   = [f for f in benchmarks if matches(f, phipatterns)]

    if any((len(testfiles)  != 2,
            len(benchmarks) != 2,
            len(testtheta)  != 1,
            len(testphi)    != 1,
            len(benchtheta) != 1,
            len(benchphi)   != 1)):
        raise ValueError('Wrong number of files: {} <-> {}'
                         ''.format(testfiles, benchmarks))

    testtheta  = pyf.imageCache[testtheta[ 0]].get_data()
    testphi    = pyf.imageCache[testphi[   0]].get_data()
    benchtheta = pyf.imageCache[benchtheta[0]].get_data()
    benchphi   = pyf.imageCache[benchphi[  0]].get_data()

    if any((testphi   .shape != testtheta.shape,
            benchtheta.shape != testtheta.shape,
            benchphi  .shape != testtheta.shape)):
        raise ValueError('Image shapes don\'t match')

    testvec  = np.zeros(tuple(testtheta.shape) + (3,))
    benchvec = np.zeros(tuple(testtheta.shape) + (3,))

    testvec[ ..., 0] = np.sin(testtheta)  * np.cos(testphi)
    testvec[ ..., 1] = np.sin(testtheta)  * np.sin(testphi)
    testvec[ ..., 2] = np.cos(testtheta)
    benchvec[..., 0] = np.sin(benchtheta) * np.cos(benchphi)
    benchvec[..., 1] = np.sin(benchtheta) * np.sin(benchphi)
    benchvec[..., 2] = np.cos(benchtheta)

    testvec  = nib.Nifti1Image(testvec,  np.eye(4), None)
    benchvec = nib.Nifti1Image(benchvec, np.eye(4), None)

    name = 'evalPolarCoordinateImageGroup_' + '_'.join(testfiles)

    pyf.imageCache[name + '_test']      = testvec
    pyf.imageCache[name + '_benchmark'] = benchvec

    return evalVectorImage(pyf, name + '_test', name + '_benchmark')


def evalGiftiVertices(pyf, testfile, benchmark):
    """Compare the vertices of two GIFTI surfaces. Uses the :func:`cmpArrays`
    function.
    """

    surf1 = pyf.imageCache[testfile]
    surf2 = pyf.imageCache[benchmark]

    # NIFTI_INTENT_POINTSET == 1008
    verts1 = [d for d in surf1.darrays if d.intent == 1008][0].data
    verts2 = [d for d in surf2.darrays if d.intent == 1008][0].data

    return cmpArrays(verts1, verts2)


def cmpArrays(arr1, arr2):
    """Compares the values in the given ``numpy`` arrays.

    Returns the mean difference between the two arrays, normalised
    by the combined data range of the two arrays.
    """

    arr1 = np.asarray(arr1, dtype=np.float128)
    arr2 = np.asarray(arr2, dtype=np.float128)

    # Non-finite values (nan/inf)
    # must match in both arrays.
    finite = np.isfinite(arr1)

    if not np.all(finite == np.isfinite(arr2)):
        return 1

    # only consider voxels where
    # we have at least one non-zero
    # value
    nzmask = finite & ((arr1 != 0) | (arr2 != 0))

    # all infinite or
    # all zero - all good
    if ~np.any(nzmask):
        return 0

    arr1 = arr1[nzmask]
    arr2 = arr2[nzmask]

    # Calculate the difference between the two
    # arrays, and scale it by the combined data
    # range of the two arrays, so that all of
    # the errors are proportional to the range
    # of the data.
    denom = np.max((arr1, arr2)) - np.min((arr1, arr2))

    # if the above results in 0, it
    # means that all voxels from both
    # arrays have the same value
    if denom == 0:
        return 0

    normdiff = np.abs((arr2 - arr1) / denom)

    # The final error is the mean error across all voxels
    return normdiff.mean()



#!/usr/bin/env python
#
# test_hashing.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


from collections import OrderedDict
import textwrap as tw
import datetime
import os.path as op
import time

from pyfeeds import hashing, testing

from . import CaptureStdout, makepyfeeds, maketest, makepaths, tempdir


def test_genHashes():

    with tempdir():
        maketest('test1/feedsRun', inputs=['test1/file1', 'test1/file2'])
        maketest('test2/feedsRun', inputs=['test2/file1', 'test2/file2'])
        makepaths(['data/test1/file1',
                   'data/test1/file2',
                   'data/test2/file1',
                   'data/test2/file2'])

        pyf = makepyfeeds(command='genhash',
                          testDir=['.'],
                          inputDir='data',
                          hashFile='hashes')
        hashing.genHashes(pyf)
        assert op.exists('hashes')

        cap = CaptureStdout()
        with cap:
            pyf = makepyfeeds(command='genhash',
                              testDir=['.'],
                              inputDir='data')
            hashing.genHashes(pyf)
            test1, test2 = pyf.tests
        print(cap.stdout)
        lines = cap.stdout.strip().split('\n')

        assert 'test1' in lines[4]                    and \
                test1.hashTestDir()       in lines[4] and \
                test1.hashInputs('data')  in lines[4]
        assert 'test2' in lines[5]                    and \
                test2.hashTestDir()       in lines[5] and \
                test2.hashInputs('data')  in lines[5]



def test_checkHashes():
    cap = CaptureStdout()
    with tempdir(), cap:
        maketest('test1/feedsRun', inputs=['test1/file1', 'test1/file2'])
        maketest('test2/feedsRun', inputs=['test2/file1', 'test2/file2'])
        makepaths(['data/test1/file1',
                   'data/test1/file2',
                   'data/test2/file1',
                   'data/test2/file2'])

        pyf = makepyfeeds(command='checkhash',
                          testDir=['.'],
                          inputDir='data')

        hashing.checkHashes(pyf)


def test_calcHashes():
    with tempdir():
        maketest('test1/feedsRun', inputs=['data/test1/file1', 'data/test1/file2'])
        maketest('test2/feedsRun', inputs=['data/test2/file1', 'data/test2/file2'])
        makepaths(['data/test1/file1',
                   'data/test1/file2',
                   'data/test2/file1',
                   'data/test2/file2'])

        test1 = testing.Test('test1/feedsRun', '.')
        test2 = testing.Test('test2/feedsRun', '.')

        testHashes, inputHashes = hashing.calcHashes([test1, test2], 'data')

        assert testHashes  == [test1.hashTestDir(),
                               test2.hashTestDir()]
        assert inputHashes == [test1.hashInputs('data'),
                               test2.hashInputs('data')]

        testHashes, inputHashes = hashing.calcHashes([test1, test2], None)
        assert testHashes  == [test1.hashTestDir(),
                               test2.hashTestDir()]
        assert inputHashes == [None, None]



def test_loadHashes():
    contents = tw.dedent("""
    123 456
    abc def

    987 654
    badline
    111 None
    """).strip()
    with tempdir():
        with open('hashfile', 'wt') as f:
            f.write(contents)

        hashes = hashing.loadHashes('hashfile')
        assert hashes == {'123' : '456',
                          'abc' : 'def',
                          '987' : '654',
                          '111' : None}


def test_saveHashes():
    contents = tw.dedent("""
    123 456
    abc def
    987 654
    111 None
    """).strip()

    hashes = OrderedDict([('123', '456'),
                          ('abc', 'def'),
                          ('987', '654'),
                          ('111', None)])

    with tempdir():

        hashing.saveHashes('hashfile', hashes)
        with open('hashfile', 'rt') as f:
            assert f.read().strip() == contents

    pass


def test_checkTestHash():
    before = datetime.datetime.now()
    time.sleep(1)
    with tempdir():
        maketest('test1/feedsRun', inputs=['test1/file1', 'test1/file2'])
        makepaths(['data/test1/file1',
                   'data/test1/file2'])

        after = datetime.datetime.now()

        pyf = makepyfeeds(command='checkhash',
                          testDir=['.'],
                          inputDir='data')
        pyf.hashesLastGenerated = before
        test = pyf.tests[0]

        hashes = { test.hashTestDir() : test.hashInputs('data') }

        assert hashing.checkTestHash(pyf, test, 'data', hashes, False, False) == hashing.TEST_HASH_PASSED

        pyf.hashesLastGenerated = after
        assert hashing.checkTestHash(pyf, test, 'data', hashes, False, False) == hashing.TEST_HASH_SKIPPED_NOCHANGE

        pyf.hashesLastGenerated = before
        hashes = { test.hashTestDir() : 'poop' }
        assert hashing.checkTestHash(pyf, test, 'data', hashes, False, False) == hashing.TEST_HASH_FAILED_DATA_HASH

        assert hashing.checkTestHash(pyf, test, 'data', {}, False, False) == hashing.TEST_HASH_FAILED_TEST_HASH



#!/usr/bin/env python
#
# imagecache.py - A cache for NIFTI images
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provide the :class:`ImageCache` class, a simple in-memory
cache for nibabel-loaded files.
"""


import            collections
import            logging
import os.path as op

import numpy   as np
import nibabel as nib


log = logging.getLogger(__name__)


def size(img):
    """Return the size of the given ``nibabel`` NIfTI image in megabytes.

    The returned size only takes into account the image data, not its header.

    Only spatial images (e.g. NIfTI) are considered; all other file types
    are assumed to take up one megabyte.
    """

    if not isinstance(img, nib.AnalyzeImage):
        return 1

    dtype = img.get_data_dtype()
    kb    = np.prod(img.shape) * dtype.itemsize
    return kb / 1048576.0


class ImageCache(object):
    """The ``ImageCache`` is just a cache for loading ``nibabel`` images (and
    other data types).  Its purpose is to avoid re-loading the same image
    multiple times - once an image has been loaded through the cache, it will
    be kept in the cache for subsequent access.

    An image can be loaded by accessing the ``ImageCache`` as a ``dict``::

        cache = ImageCache()

        # the image is loaded on first access
        img = cache['myimage.nii.gz']

        # on subsequent accesses, the image
        # is returned from the cache
        img = cache['myimage.nii.gz']
    """

    def __init__(self, maxsize=32768):
        """Create the cache.

        :arg maxsize: Maximum size of the cache in megabytes. When the cache
                      reaches this size, old entries are dropped.
        """
        self.__images      = collections.OrderedDict()
        self.__imagesizes  = {}
        self.__currentsize = 0
        self.__maxsize     = maxsize

        log.debug('Initialising ImageCache(maxsize=%iGB)', maxsize / 1024)


    def __getitem__(self, imagefile):
        """Return the ``nibabel`` image corresponding to the given
        ``imagefile``, loading it if necessary.
        """

        imagefile = op.realpath(op.abspath(imagefile))
        image     = self.__images.get(imagefile, None)

        if image is not None:

            # keep the images dict
            # ordered by access time
            log.debug('Touching %s', imagefile)
            self.__images.move_to_end(imagefile)
            return image

        image           = nib.load(imagefile)
        self[imagefile] = image
        return image


    def __setitem__(self, imagefile, image):
        """Add something to the cache. This is intended to be used for
        storing derived, in-memory-only images.
        """

        imagesize                    = size(image)
        self.__images[    imagefile] = image
        self.__imagesizes[imagefile] = imagesize

        newsize = self.__currentsize + imagesize

        log.debug('Adding %s (%0.2f / %0.0f)',
                  imagefile, newsize, self.__maxsize)

        while newsize > self.__maxsize:
            lrufile  = self.__images    .popitem(last=False)[0]
            lrusize  = self.__imagesizes.pop(    lrufile)

            log.debug('Removing %s (%0.2f)', lrufile, lrusize)

            newsize -= lrusize

        self.__currentsize = newsize



#!/usr/bin/env python
#
# testing.py - the testing module
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``testing`` module contains the :class:`Test` class, which represents
a single pyfeeds test. It also contains the following functions:

.. autosummary::
   :nosignatures:

   findTestDirs
   findTestScripts
   listTests
   bundleTests
   runTests
   runTest
"""


from __future__ import print_function
from __future__ import division

import logging

import            os
import os.path as op
import            re
import            sys
import            glob
import            time
import            shutil
import            hashlib
import            subprocess

from multiprocessing.pool import ThreadPool as Pool

from . import evaluate
from . import hashing
from . import common


log = logging.getLogger(__name__)



SCRIPTFILE_PATTERN = r'feedsRun(\.\S+)?'
"""Test scripts must have a name which matches this regular expression.
"""


class Test(object):
    """The ``Test`` class represents a single pyfeeds test. A ``Test`` object
    has a few useful methods:


    .. autosummary::
       :nosignatures:

       hashInputs
       hashTestDir
       createSandbox


    The following attributes are available on a ``Test`` object:

    ============== ==========================================================
    ``name``       A unique, human readable, internal name for this ``Test``,
                   generated from the test directory path.

    ``testDir``    The test directory.

    ``testScript`` The full path to the ``feedsRun`` test executable.
    ============== ==========================================================
    """


    def __init__(self, testScript, baseDir):
        """Create a ``Test`` object.

        :arg testScript: Path to the test script.

        :arg baseDir:    The *base* directory. This is identified by pyfeeds as
                         the parent directory of all tests which have been
                         specified by the user - see the
                         :func:`.nearestCommonAncestor` function.
        """

        testScript = op.abspath( testScript)
        fileName   = op.basename(testScript)

        if not op.isfile(testScript):
            raise RuntimeError('Test script {} does not '
                               'exist'.format(testScript))
        if not re.fullmatch(SCRIPTFILE_PATTERN, fileName):
            raise RuntimeError('Invalid test script name: '
                               '{}'.format(testScript))

        suffix  = fileName[9:]
        testDir = op.dirname( testScript)
        baseDir = op.abspath( baseDir)

        if testDir == baseDir:
            name = op.basename(testDir)
        else:
            name = testDir[len(baseDir):].strip(op.sep)

        if len(suffix) > 0:
            name = op.sep.join((name, suffix))

        self.name       = name
        self.testDir    = testDir
        self.baseDir    = baseDir
        self.testScript = testScript
        self.__inputs   = self.__loadInputsFile()


    def __str__(self):
        """Returns a string representation of this ``Test``."""
        return self.name


    def __repr__(self):
        """Returns a string representation of this ``Test``."""
        return self.name


    def getInputs(self):
        """Return a list containing the paths to the external data
        dependencies of this ``Test``.
        """
        return list(self.__inputs)


    def inputsChangedSince(self, inputDir, mtime):
        """Returns ``True`` if any of the input files for this test
        have changed since the given ``datetime`` object.
        """
        inputs = [op.join(inputDir, i) for i in self.__inputs]
        return not all([common.allBefore(i, mtime) for i in inputs])


    def hashTestDir(self):
        """Calculates and returns MD5 digest of the contents of the test
        directory. See the :func:`.hashPath` function.
        """
        log.debug('Test {} - hashing test directory'.format(self.name))

        hashObj = hashlib.md5()
        count   = hashing.hashPath(self.testDir, hashObj)
        digest  = hashObj.hexdigest()

        log.debug('Test {} - test directory hash ({} files): {}'.format(
            self.name, count, digest))

        return digest


    def hashInputs(self, inputDir):
        """Calculates and returns MD5 digest of all external/shared data upon
        which this ``Test`` is dependent. See the :func:`.hashPath` function.

        :arg inputDir: Directory in which the test input data is located.
        """

        log.debug('Test {} - hashing inputs'.format(self.name))

        hashObj = hashlib.md5()

        count = 0
        for src in self.__inputs:
            src     = op.join(inputDir, src)
            count  += hashing.hashPath(src, hashObj)

        if count == 0: digest = None
        else:          digest = hashObj.hexdigest()

        log.debug('Test {} - input hash ({} files): {}'.format(
            self.name, count, digest))

        return digest


    def createSandbox(self, inputDir, sandbox=None, symlink=True):
        """Creates a *sandbox* directory for this ``Test``.  The resulting
        sandbox directory contains symlinks or copies of all the
        external/shared data dependencies of this ``Test``.

        See the :func:`.common.createSandbox` function for details on the
        arguments.

        :returns:         If this ``Test`` has no external data dependencies,
                          ``None`` is returned. Otherwise, the path to the
                          sandbox directory is returned.
        """

        return common.createSandbox(inputDir,
                                    self.__inputs,
                                    sandbox=sandbox,
                                    symlink=symlink)


    def __loadInputsFile(self):
        """Loads the all of the paths contained in the ``feedsInputs`` file,
        contained in this ``Test`` directory. Returns a list containing all
        of the paths. The returned list will be empty if this ``Test`` has no
        external data dependencies.
        """

        inputsFile = op.join(self.testDir, 'feedsInputs')

        if op.isfile(inputsFile):

            log.debug('Test {} - reading inputs from {}'.format(
                self.name, inputsFile))

            with open(inputsFile, 'rt') as f:

                # Make sure leading/trailing
                # whitespace and slashes, and
                # blank lines are removed
                inputs = list(f)
                inputs = [i.strip()       for i in inputs]
                inputs = [i.strip(op.sep) for i in inputs]
                inputs = [i               for i in inputs if i != '']
        else:
            inputs = []

        inputs = list(sorted(set(inputs)))

        return inputs


def findTestDirs(path, depth=0):
    """Recursively finds all sub-directories in the given path which
    look like they contain a pyfeeds test.
    """

    path     = op.abspath(path)
    space    = ' ' * depth
    testDirs = []

    if depth == 0: logfunc = log.info
    else:          logfunc = log.debug

    logfunc('{}Searching for tests in {} ...'.format(space, path))

    # is this directory a test directory?
    if len(findTestScripts(path)) > 0:
        log.debug('{}Found test directory: {}'.format(space, path))
        testDirs.append(path)

    # Gather a list of all sub-directories
    subDirs = [op.join(path, subDir) for subDir in os.listdir(path)]
    subDirs = [sd for sd in subDirs if op.isdir(sd)]

    # Recursively search for tests in the
    # sub-directories of this directory
    for sd in subDirs:
        testDirs.extend(findTestDirs(sd, depth + 1))

    return testDirs


def findTestScripts(testDir):
    """Return a list containing all ``feedsRun`` scripts in ``testDir``.

    :arg testDir: Directory which may contain ``feedsRun`` scripts.
    """
    pattern = SCRIPTFILE_PATTERN
    hits    = [              op.join(testDir, 'feedsRun')]
    hits   += list(glob.glob(op.join(testDir, 'feedsRun.*')))
    hits    = [h for h in hits if re.fullmatch(pattern, op.basename(h))]
    hits    = [h for h in hits if op.isfile(h)]

    return hits


def listTests(pyf):
    """Prints a list of all available tests.

    :arg pyf: A :class:`.Pyfeeds` instance
    """

    testNames = [t.name    for t in pyf.tests]
    testDirs  = [t.testDir for t in pyf.tests]
    titles    = ['Test name', 'Test directory']

    print('\nAvailable tests\n')
    common.printColumns(titles, [testNames, testDirs])


def bundleTests(pyf):
    """Creates a test *bundle* containing all of the given tests.

    :arg pyf: A :class:`.Pyfeeds` instance.
    """

    tests = pyf.tests

    bundleDir       = op.abspath(pyf.outputDir)
    inputDir        =            pyf.inputDir
    benchmarkDir    =            pyf.benchmarkDir
    bunTestDir      = op.join(bundleDir, 'tests')
    bunInputDir     = op.join(bundleDir, 'data')
    bunBenchmarkDir = op.join(bundleDir, 'benchmarks')

    if inputDir     is not None: inputDir     = op.abspath(inputDir)
    if benchmarkDir is not None: benchmarkDir = op.abspath(benchmarkDir)

    if op.exists(bundleDir) and pyf.overwrite:
        shutil.rmtree(bundleDir)

    if op.exists(bundleDir):
        log.error('Bundle directory already exists ({})!'.format(bundleDir))
        sys.exit(1)

    os.makedirs(bunTestDir)

    if inputDir     is not None: os.makedirs(bunInputDir)
    if benchmarkDir is not None: os.makedirs(bunBenchmarkDir)

    # We build a list of the inputs
    # for every test, and then pass
    # them to the common.createSandbox
    # function afterwards.
    inputs = []

    log.info('Copying tests to bundle ...')

    for i, test in enumerate(tests):

        destDir = op.join(bunTestDir, test.name)

        inputs.extend(test.getInputs())

        if op.exists(destDir):
            continue

        log.debug('Copying test directory {} to bundle '
                  '(tests/{})'.format(test.testDir, test.name))

        shutil.copytree(test.testDir, destDir)

    # Copy the test input data
    if inputDir is not None:
        log.info('Copying shared test data to bundle ...')
        common.createSandbox(inputDir,
                             inputs,
                             sandbox=bunInputDir,
                             symlink=False)

    # Copy the test benchmark data
    if benchmarkDir is not None:
        log.info('Copying test benchmark data to bundle ...')

        for i, test in enumerate(tests):

            testBenchmarkDir = common.findPathIn(test.testDir, benchmarkDir)

            if testBenchmarkDir is None:
                log.debug('Could not find any benchmark '
                          'data for test {}'.format(test.name))
                continue

            destRelDir = op.relpath(testBenchmarkDir, benchmarkDir)
            destDir   = op.join(bunBenchmarkDir, destRelDir)
            if op.exists(destDir):
                continue

            log.debug('Copying test benchmark directory {} to bundle '
                      '(benchmarks/{})'.format(testBenchmarkDir, destRelDir))

            parentDir = op.split(destDir)[0]

            if not op.exists(parentDir):
                os.makedirs(parentDir)

            shutil.copytree(testBenchmarkDir, destDir)

    # Create a .feedsHashes
    # file if requested
    if (inputDir is not None) and pyf.genHashes:

        hashFile                = op.join(bunInputDir, '.feedsHashes')
        testHashes, inputHashes = hashing.calcHashes(tests, bunInputDir)

        hashes = {th : ih for th, ih in zip(testHashes, inputHashes)}

        hashing.saveHashes(hashFile, hashes)


def runTests(pyf):
    """Runs all of the given tests - see the :func:`runTest` function.

    :arg pyf: A :class:`.Pyfeeds` instance.
    """

    tests = pyf.tests

    outputDir    = op.abspath(pyf.outputDir)
    inputDir     =            pyf.inputDir
    benchmarkDir =            pyf.benchmarkDir
    hashFile     =            pyf.hashFile
    jobPool      = Pool(      pyf.jobs)
    startTime    = time.time()

    if inputDir     is not None: inputDir     = op.abspath(inputDir)
    if benchmarkDir is not None: benchmarkDir = op.abspath(benchmarkDir)

    if op.exists(outputDir) and pyf.overwrite:
        shutil.rmtree(outputDir)

    if op.exists(outputDir):
        log.error('Output directory exists ({})!'.format(outputDir))
        sys.exit(1)

    if (pyf.sandboxDir is not None) and op.exists(pyf.sandboxDir):
        log.error('Sandbox directory exists ({})!'.format(outputDir))
        sys.exit(1)

    # Don't update hashes if
    # we're skipping hashes
    pyf.updateHashes &= (not pyf.skipHashes)

    if (hashFile is not None) and \
       (inputDir is not None) and \
       (not pyf.skipHashes):
        hashFile   = op.join(inputDir, '.feedsHashes')
        hashFile   = op.abspath(hashFile)
        testHashes = hashing.loadHashes(hashFile)
    else:
        testHashes = {}

    log.info('{} tests to run'.format(len(tests)))

    # Make sure that the pyfeeds directory
    # is on the python path, so that individual
    # tests have access to the pyfeeds package
    # namespace if they wish
    thisDir           = op.join(op.dirname(op.realpath(__file__)), '..')
    env               = os.environ.copy()
    env['PYTHONPATH'] = op.pathsep.join([thisDir, env.get('PYTHONPATH', '')])

    # For each test we store a dictionary containing
    # the results of running the test. Each result
    # dict contains:
    #
    #   - 'test'      - The Test object
    #   - 'hash_code' - Test hash code (see hashing.checkTestHash)
    #   - 'exit_code' - Test process exit code
    #   - 'time'      - Number of seconds taken to run the test
    #
    # These are then passed through to the
    # evaluate.evaluateTest function, to
    # evaluate the test.
    testResults = [{} for t in tests]
    for i, test in enumerate(tests):
        testResults[i]['test']      = test
        testResults[i]['hash_code'] = hashing.TEST_HASH_NOT_CHECKED

    # Verify the input data
    # hash for every test
    if not pyf.skipHashes:

        # Do it in parallel using the job pool.
        jobArgs = [(pyf, t, inputDir, testHashes,
                    pyf.updateHashes, pyf.forceHashes)
                   for t in tests]
        results = jobPool.map(common.starmap(hashing.checkTestHash), jobArgs)

        # Update the test result dicts
        for i, (test, result) in enumerate(zip(tests, results)):
            testResults[i]['hash_code'] = result

    # Run each test. Build a list of
    # arguments to pass to the runTest
    # function for every test.
    jobArgs = []
    for test in tests:

        if pyf.sandboxDir is None:
            sandbox = None
        else:

            sandbox = op.abspath(op.join(pyf.sandboxDir, test.name))

            if op.exists(sandbox):
                log.debug('Removing old sandbox '
                          'directory {}'.format(sandbox))
                shutil.rmtree(sandbox)

        jobArgs.append((
            pyf,
            test,
            outputDir,
            inputDir,
            benchmarkDir,
            env,
            pyf.leaveSandboxes,
            sandbox))

    # Then run the tests
    # using the job pool.
    results = jobPool.map(common.starmap(runTest), jobArgs)
    endTime = time.time()
    minutes = (endTime - startTime) / 60.0

    # Copy the test results and runtimes
    # to the TestResult objects
    for i, (test, (result, runtime)) in enumerate(zip(tests, results)):

        testResults[i]['result'] = result
        testResults[i]['time']   = runtime

    # Then arrange the test result
    # info into separate columns
    titles  = ['Test', 'Time (seconds)', 'Result', 'Data status']
    columns = [
        [r['test'].name                                  for r in testResults],
        [r['time']                                       for r in testResults],
        ['Passed' if r['result'] else 'Failed'           for r in testResults],
        [hashing.TEST_HASH_SHORT_STRINGS[r['hash_code']] for r in testResults]]

    # And print them all out nicely
    common.printColumns(titles, columns)

    passed = sum([r['result'] for r in testResults])
    log.info('{} / {} tests passed in {:0.2f} minutes'.format(
        passed, len(testResults), minutes))

    if pyf.updateHashes:
        hashing.saveHashes(hashFile, testHashes)


def runTest(pyf,
            test,
            outputDir,
            inputDir,
            benchmarkDir,
            env,
            leaveSandbox,
            sandboxDir=None):
    """Runs the given :class:`Test`. The test is executed, and then passed to
    the :func:`.evaluateTest` function to determine whether it passed or
    failed.

    :returns: A tuple containing:

                - ``True`` if the test passed, ``False`` otherwise.
                - Number of seconds taken to run the test

    :arg pyf:          The :class:`.Pyfeeds` instance.

    :arg test:         The test.

    :arg outputDir:    Output directory.

    :arg inputDir:     Directory containing test input data.

    :arg benchmarkDir: Directory containing test benchmark data.

    :arg env:          Dictionary containing the environment variables to be
                       passed to the test process.

    :arg leaveSandbox: If ``True``, the sandbox directory created for the test
                       (if any) is not deleted.

    :arg sandboxDir:   Directory to be used as the test sandbox. See the
                       ``sandbox`` parameter to :meth:`Test.createSandbox`.
    """

    runTime = 0

    try:

        outputDir = op.realpath(op.join(outputDir, test.name))

        if inputDir is not None:
            sandboxDir = test.createSandbox(inputDir, sandbox=sandboxDir)

        if benchmarkDir is not None:
            benchmarkDir = common.findPathIn(test.testDir, benchmarkDir)

        log.info( 'Running test {} ...'.format(test.name))
        log.debug('   input directory:     {}'.format(sandboxDir))
        log.debug('   output directory:    {}'.format(outputDir))
        log.debug('   benchmark directory: {}'.format(benchmarkDir))

        if op.isdir(outputDir):

            log.debug('Deleting old test output directory: {}'.format(
                outputDir))
            shutil.rmtree(outputDir)

        log.debug('(Re-)Creating test output directory: {}'.format(outputDir))
        os.makedirs(outputDir)

        cmd = [test.testScript, outputDir, sandboxDir, benchmarkDir]
        cmd = [str(c) for c in cmd]

        start = time.time()

        with open(op.join(outputDir, 'feedsRun.log'), 'wt') as logFile:

            log.debug('Calling process "{}"'.format(' '.join(cmd)))

            retcode = subprocess.call(cmd,
                                      env=env,
                                      cwd=test.testDir,
                                      stdout=logFile,
                                      stderr=logFile)

        runTime = int(time.time() - start)
        result  = evaluate.evaluateTest(pyf,
                                        test,
                                        retcode,
                                        outputDir,
                                        benchmarkDir)

        if result: status = 'passed'
        else:      status = 'failed'

        log.info('Test {} {} in {} seconds'.format(test, status, str(runTime)))

        if (sandboxDir is not None) and (not leaveSandbox):
            log.debug('Deleting sandbox for test {}: {}'.format(
                test, sandboxDir))
            shutil.rmtree(sandboxDir)

        return result, runTime

    except Exception as e:

        log.warning('Test {} crashed: {}'.format(test.name, str(e)),
                    exc_info=True)

        return False, runTime



#!/usr/bin/env python
#
# common.py - Miscellaneous functions
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains some miscellaneous utility functions used throughout
pyfeeds, and made available to test scripts.

.. autosummary::
   :nosignatures:

   printColumns
   findCommonAncestors
   findPathIn
   createSandbox
   loadNumericText
   allBefore
   starmap
   fslbin
   readfsf
   writefsf
"""


from __future__ import print_function
from __future__ import division

import            collections
import            datetime
import            tempfile
import            logging
import            shutil
import            re
import            os
import os.path as op
import numpy   as np


log = logging.getLogger(__name__)


def printColumns(titles, columns):
    """Convenience function which pretty-prints a collection of columns in a
    tabular format.

    :arg titles:  A list of titles, one for each column.

    :arg columns: A list of columns, where each column is a list of strings.
    """

    cols  = []

    for t, c in zip(titles, columns):
        cols.append([t] + list(map(str, c)))

    columns = cols
    colLens = []

    for col in columns:
        maxLen = max([len(r) for r in col])
        colLens.append(maxLen)

    fmtStr = ' | '.join(['{{:<{}s}}'.format(l) for l in colLens])

    titles  = [col[0]  for col in columns]
    columns = [col[1:] for col in columns]

    separator = ['-' * l for l in colLens]

    print(fmtStr.format(*titles))
    print(fmtStr.format(*separator))

    nrows = len(columns[0])
    for i in range(nrows):

        row = [col[i] for col in columns]
        print(fmtStr.format(*row))



def findCommonAncestor(paths):
    """Finds the nearest common ancestor, in the file system hierarchy, of
    all the given paths.
    """

    pathLens  = [len(p) for p in paths]
    template  = paths[np.argmin(pathLens)]
    parents   = template.split(op.sep)[:-1]
    depth     = len(parents)

    while depth > 0:

        path = op.sep.join(parents[:depth])

        for p in paths:
            if not p.startswith(path):
                break
        else:
            break

        depth -= 1

    return op.sep.join(parents[:depth])


def findPathIn(path, base):
    """Searches for the given ``path`` in the given ``base`` directory.

    The search begins with the full ``path``. If ``base/path`` does not
    exist, the leading directory from ``path`` is removed, and the search
    repeated.

    Returns the full path to ``base/path``, or ``None`` if the ``path`` cannot
    be found in ``base``.

    This function is used to search for a benchmark data directory which
    corresponds to a specific test directory. The benchmark data directory
    must have the same folder structure as the base directory of all of the
    tests that pyfeeds is given. But pyfeeds does not necessarily know the
    base test directory. So when pyfeeds needs to find the benchmark data for
    a given test, it uses this function to find it.

    This is easier to explain with an example. Say our benchmark data is
    stored in::

        /opt/pyfeeds/benchmarkData/

    And we want to find the benchmark data for a test stored in::

        /opt/pyfeeds/testSuite/reg/flirt/

    Which is stored in::

        /opt/pyfeeds/benchmarkData/reg/flirt

    This function will iteratively test the following paths until it finds a
    match::

        /opt/pyfeeds/benchmarkData/opt/pyfeeds/testSuite/reg/flirt/
        /opt/pyfeeds/benchmarkData/pyfeeds/testSuite/reg/flirt/
        /opt/pyfeeds/benchmarkData/testSuite/reg/flirt/
        /opt/pyfeeds/benchmarkData/reg/flirt/

    Obviously this is not a foolproof method, but the current design of
    pyfeeds disallows test directories to be unambiguously matched to
    benchmark directories.
    """

    if path.startswith(op.sep):
        path = path[1:]

    while len(path) > 0:

        test = op.join(base, path)

        if op.exists(test):
            return test

        path = op.sep.join(path.split(op.sep)[1:])

    return None


def createSandbox(inputDir, inputs, sandbox=None, symlink=True):
    """Creates a *sandbox* directory which contains symlinks or copies of all
    files in the given list of ``inputs``.

    :arg inputDir: Directory in which the data is located.

    :arg inputs:   A list of paths, specified relative to the ``inputDir``,
                   which should be copied / symlinked into the sandbox.

    :arg sandbox:  Directory in which the sandbox should be created. If
                   ``None``, a temporary directory is created and used.
                   It is up to the caller to remove this directory if
                   necessary.

    :arg symlink:  If ``True`` (the default), the sandbox is created
                   with symlinks to the input data. Otherwise, the
                   input data is copied to the sandbox directory.

    :returns:      The path to the sandbox directory, or ``None`` if
                   ``inputs`` is empty.
    """

    if len(inputs) == 0:
        return None

    if sandbox is None: sandbox = tempfile.mkdtemp(prefix='pyfeeds_')
    else:               sandbox = op.abspath(sandbox)

    log.debug('Creating sandbox: {}'.format(sandbox))

    # We sort input alphabetically,
    # and keep keep track of paths
    # that have already been copied.
    # Sorting ensures that if we have
    # two paths like this:
    #
    # /inputDir/testData/
    # /inputDir/testData/file.nii.gz
    #
    # They will always be ordered such
    # that the shorter one will come
    # first. The code below depends on
    # this ordering.
    inputs = sorted(set(inputs))
    copied = set()

    for src in inputs:

        log.debug('Processing input: {}'.format(src))

        relSrc  = src
        srcName = op.basename(src)
        src     = op.join(inputDir, src)

        if not op.exists(src):
            log.warning('Invalid path: {}'.format(src))
            continue

        # Has the directory in which this input
        # is contained already been copied /
        # linked?
        #
        # Test to see if any of the parent
        # directories of the input have been
        # added to the copied set.
        path          = src
        alreadyCopied = False
        while len(path) != 0:

            path = path.rstrip(op.sep)

            if path in copied:
                alreadyCopied = True
                break

            path = op.dirname(path)

        if alreadyCopied:
            log.debug('Skipping duplicate path: {}'.format(src))
            continue

        # Keep track of all copied / linked files / directoriess
        copied.add(src)

        relSrcDir = op.dirname(relSrc)
        destDir   = op.join(sandbox,  relSrcDir)
        dest      = op.join(destDir,  srcName)

        if not op.exists(destDir):
            os.makedirs(destDir)

        if symlink:
            log.debug('Sym-linking: {} -> {}'.format(src, dest))
            os.symlink(src, dest)

        else:
            log.debug('Copying: {} -> {}'.format(src, dest))

            if op.isdir(src): shutil.copytree(src, dest)
            else:             shutil.copy(    src, dest)

    return sandbox


def loadNumericText(filename):
    """Attempts to load any and all numbers from the given plain text file.
    Returns all found numbers as a ``numpy`` array.
    """

    # http://perldoc.perl.org/perlretut.html#Building-a-regexp
    # http://stackoverflow.com/a/385597
    numberPattern = r'[+-]? *(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?'
    numberPattern = re.compile(numberPattern)

    def stof(s):
        try:               return float(s)
        except ValueError: return np.nan

    with open(filename, 'rt') as f:
        text    = f.read()
        numbers = numberPattern.findall(text)
        numbers = [stof(n) for n in numbers]

        return np.array(numbers)


def allBefore(path, timeThres):
    """Recursively checks the last-modified time of every file in the given
    ``path`` against the given ``timeThres``.

    :returns:       ``True`` if no files were modified after the
                    ``timeThres``, ``False`` otherwise.

    :arg path:      Path to a directory or file.

    :arg timeThres: ``datatime`` object containing the last modified
                    threshold.
    """

    mtime = datetime.datetime.fromtimestamp(op.getmtime(path))

    if mtime >= timeThres:
        return False

    if op.isdir(path):
        for p in os.listdir(path):
            if not allBefore(op.join(path, p), timeThres):
                return False

    return True


def starmap(func):
    """This function is used as a decorator on functions which
    are passed to the ``multiprocessing.Pool.map`` function, to allow
    us to pass un-expanded argument lists to the function being
    executed.

    .. note:: This function is part of the standard Python library
              in Python 3.3 and newer, but does not exist in older
              versions.
    """
    def wrapper(args):
        return func(*args)
    return wrapper


def fslbin(name):
    """Returns an absolute path to the named FSL command. """
    return op.join(os.environ["FSLDIR"], "bin", name)


def readfsf(filename, keys):
    """Reads the given ``keys`` from a FEAT ``fsf`` configuration file. """
    if not isinstance(keys, collections.Sequence):
        keys = [keys]

    values = {}

    with open(filename, 'rt') as f:
        for line in f:

            if not line.startswith('set '):
                continue
            line = line.split()
            k    = line[1]
            v    = ' '.join(line[2:])
            if k in keys:
                values[k] = v
    return [values[key] for key in keys]


def writefsf(filename, keys, values):
    """Writes the given ``keys`` and corresponding ``values`` to a FEAT
    ``fsf`` configuration file.
    """

    if not isinstance(keys,   collections.Sequence): keys   = [keys]
    if not isinstance(values, collections.Sequence): values = [values]

    if len(keys) != len(values):
        raise ValueError('len(keys) != len(values)')

    with open(filename, 'at') as f:
        for k, v in zip(keys, values):
            f.write('\nset {} {}'.format(k, v))
